# -*- coding: utf-8 -*-

class QC:
    FAIL = 'FAIL'
    WARNING = 'WARNING'
    PASS = 'PASS'
    AMBIGUOUS_RESULTS_ERROR_3 = 'Ambiguous Results Error 3'
    UNCONFIDENT_RESULTS_ERROR_4 = 'Unconfident Results Error 4'
    NO_SUBTYPE_RESULT = 'No subtype result!'
    NO_TARGETS_FOUND = 'No tiles/targets were found in this sample.'
